@@ -1,7 +1,7 @@
 /*
  * pam_tally.c
  * 
- * $Id: pam_tally.c,v 1.9 2001/12/09 21:44:58 agmorgan Exp $
+ * $Id: pam_tally.c,v 1.11 2004/09/24 13:13:21 kukuk Exp $
  */
 
 
@@ -49,6 +49,7 @@
 /* #define PAM_SM_PASSWORD */
 
 #include <security/pam_modules.h>
+#include <security/_pam_modutil.h>
 
 /*---------------------------------------------------------------------*/
 
@@ -109,7 +110,7 @@ static char *cline_user=0;  /* cline_user is used in the administration prog */
 
 static int pam_get_uid( pam_handle_t *pamh, uid_t *uid, const char **userp ) 
   {
-    const char *user;
+    const char *user = NULL;
     struct passwd *pw;
 
 #ifdef MAIN
@@ -123,7 +124,7 @@ static int pam_get_uid( pam_handle_t *pamh, uid_t *uid, const char **userp )
       return PAM_AUTH_ERR;
     }
 
-    if ( ! ( pw = getpwnam( user ) ) ) {
+    if ( ! ( pw = _pammodutil_getpwnam( pamh, user ) ) ) {
       _pam_log(LOG_ERR,MODULE_NAME ": pam_get_uid; no such user %s",user);
       return PAM_USER_UNKNOWN;
     }
@@ -189,7 +190,8 @@ static int get_tally( tally_t *tally,
 
     if ( fseek( *TALLY, uid * sizeof(struct faillog), SEEK_SET ) ) {
           _pam_log(LOG_ALERT, "fseek failed %s", filename);
-                return PAM_AUTH_ERR;
+          fclose(*TALLY);
+          return PAM_AUTH_ERR;
     }
                     
     if ( fileinfo.st_size <= uid * sizeof(struct faillog) ) {
@@ -527,6 +529,8 @@ PAM_FUNCTION( pam_sm_acct_mgmt ) {
 			 user,uid,
 			 fsp->fs_fail_time+fsp->fs_faillog.fail_locktime
 			 -time(NULL));
+		if (TALLY)
+			fclose(TALLY);
       		return PAM_AUTH_ERR;
       	}
       }
@@ -537,6 +541,8 @@ PAM_FUNCTION( pam_sm_acct_mgmt ) {
         ) {
         _pam_log(LOG_NOTICE,"user %s ("UID_FMT") tally "TALLY_FMT", deny "TALLY_FMT,
                  user, uid, tally, deny);
+		if (TALLY)
+			fclose(TALLY);
         return PAM_AUTH_ERR;                 /* Only unconditional failure   */
       }
       
